#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdbool.h>
#include <poll.h>
#include <inttypes.h>
#include <unistd.h>
#include <errno.h>
#include "common.h"

#define PRU0_DEV    "/sys/class/remoteproc/remoteproc1"
#define RPMSG0      "/dev/rpmsg_pru30"
#define FW_NAME     "pru_fw.out"

// IRQSTATUS_RAW flags
#define OUT_OF_RANGE    0b100000000
#define FIFO1_UNDER_FLO 0b010000000
#define FIFO1_OVER_RUN  0b001000000
#define FIFO1_THRESH    0b000100000
#define FIFO0_UNDER_FLO 0b000010000
#define FIFO0_OVER_RUN  0b000001000
#define FIFO0_THRESH    0b000000100
#define END_OF_SEQ      0b000000010

bool load_pru_fw(const char *name){
    int st, fw;
    char buf[256];
    // stop running fw
    if((st = open(PRU0_DEV "/state", O_RDWR)) < 0){
        perror("open: " PRU0_DEV "/state");
        return(1);
    }
    if(read(st, &buf, sizeof(buf)) != 8){
        perror("read: " PRU0_DEV "/state");
        return(1);
    }
    if(!strncmp(buf, "running\n", 8)){
        if(lseek(st, 0, SEEK_SET) < 0){
            perror("lseek: " PRU0_DEV "/state");
            return(1);
        }
        if(write(st, "stop", 4) != 4){
            perror("write: " PRU0_DEV "/state");
            return(1);
        }
        if(lseek(st, 0, SEEK_SET) < 0){
            perror("lseek: " PRU0_DEV "/state");
            return(1);
        }
        if(read(st, &buf, sizeof(buf)) != 8){
            perror("read: " PRU0_DEV "/state");
            return(1);
        }
        if(strncmp(buf, "offline\n", 8)){
            fprintf(stderr, "Can not stop FW\n");
            return(1);
        }
    }

    // load fw
    if((fw = open(PRU0_DEV "/firmware", O_RDWR)) < 0){
        perror("open:" PRU0_DEV "/firmware");
        return(1);
    }
    if(write(fw, FW_NAME, strlen(FW_NAME)) != strlen(FW_NAME)){
        perror("write: " PRU0_DEV "/firmware");
        return(1);
    }
    if(lseek(fw, 0, SEEK_SET) < 0){
        perror("lseek: " PRU0_DEV "/firmware");
        return(1);
    }
    if(read(fw, &buf, sizeof(buf)) != strlen(FW_NAME)+1 ){
        perror("read: " PRU0_DEV "/firmware");
        return(1);
    }
    if(strncmp(buf, FW_NAME, strlen(FW_NAME))){
        fprintf(stderr, "Can not load FW %s\n", FW_NAME);
        return(1);
    }
    close(fw);

    // start fw
    if(write(st, "start", 5) != 5){
        perror("write: " PRU0_DEV "/state");
        return(1);
    }
    if(lseek(st, 0, SEEK_SET) < 0){
        perror("lseek: " PRU0_DEV "/state");
        return(1);
    }
    if(read(st, &buf, sizeof(buf)) != 8){
        perror("read: " PRU0_DEV "/state");
        return(1);
    }
    if(strncmp(buf, "running\n", 8)){
        fprintf(stderr, "Can not start FW\n");
        return(1);
    }

    close(st);
    return(0);
}

int main(void) {
    if(load_pru_fw(FW_NAME)){
        fprintf(stderr, "Can not start PRU FW\n");
        return EXIT_FAILURE;
    }

	shared_struct dat = {};
    pru_cmd cmd = PRU_BUSY;

    sleep(1);

    int fd = open(RPMSG0, O_RDWR);
    if(fd < 0){
        perror("open: " RPMSG0);
        return EXIT_FAILURE;
    }
    
    // Start sampling
    cmd=PRU_START;
    while(1){
        if(write(fd, &cmd, sizeof(cmd)) != sizeof(cmd)){
            perror("write: "RPMSG0);
            return EXIT_FAILURE;
        }
        if(read(fd, &dat, sizeof(dat)) != sizeof(dat)){
            perror("read: " RPMSG0);
            return EXIT_FAILURE;
        }
        if((dat.irqstatus_raw & 0b1111011000) == 0){
            int x[3]={};
            for(int i=0; i<128; i++){
	            printf("%d\n", dat.voltage[i]*1800/(1<<12));
            }
        }
        else {
            if(dat.irqstatus_raw & OUT_OF_RANGE){
	            fprintf(stderr, "Out of range. 0x%x\n", dat.irqstatus_raw);
            }
            else if(dat.irqstatus_raw & FIFO1_UNDER_FLO){
	            fprintf(stderr, "FIFO1 under flow.\n");
            }
            else if(dat.irqstatus_raw & FIFO1_OVER_RUN){
	            fprintf(stderr, "FIFO1 over run.\n");
            }
            else if(dat.irqstatus_raw & FIFO0_UNDER_FLO){
	            fprintf(stderr, "FIFO0 under flow.\n");
            }
            else if(dat.irqstatus_raw & FIFO0_OVER_RUN){
	            fprintf(stderr, "FIFO0 over run.\n");
            }
            else {
	            fprintf(stderr, "Unknown error. 0x%x\n", dat.irqstatus_raw);
            }
            return EXIT_FAILURE;
        }
    }
	return EXIT_SUCCESS;
}
